"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _ethereumjsUtil = _interopRequireDefault(require("ethereumjs-util"));

var _jsbiUtils = require("jsbi-utils");

var _nodeFetch = _interopRequireDefault(require("node-fetch"));

var _web3CorePromievent = _interopRequireDefault(require("web3-core-promievent"));

var _transaction = _interopRequireDefault(require("./transaction"));

var _outpoint = _interopRequireDefault(require("./outpoint"));

var _output = _interopRequireDefault(require("./output"));

var _input = _interopRequireDefault(require("./input"));

var _period = _interopRequireDefault(require("./period"));

var _helpers = require("./helpers");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance"); }

function _iterableToArrayLimit(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var Exit =
/*#__PURE__*/
function () {
  function Exit() {
    _classCallCheck(this, Exit);
  }

  _createClass(Exit, null, [{
    key: "signOverExit",

    /**
     * Returns:
     * 0 -  32 bytes utxoId
     * 32 -  32 bytes amount of money
     * 64 -  32 bytes r-signature
     * 96 -  32 bytes s-signature
     * 128 - 31 bytes padding
     * 159 -  1 byte v-signature
     */
    value: function signOverExit(utxoId, amount, privKey) {
      var sigHashBuff = Exit.sigHashBuff(utxoId, amount);

      var sigHash = _ethereumjsUtil["default"].hashPersonalMessage(sigHashBuff);

      var sig = _ethereumjsUtil["default"].ecsign(sigHash, Buffer.from(privKey.replace('0x', ''), 'hex'));

      var vBuff = Buffer.alloc(32);
      vBuff.writeInt8(sig.v, 31);
      return Buffer.concat([sigHashBuff, sig.r, sig.s, vBuff]);
    }
    /**
     * params:
     * utxoId: string with 0x, e.g. 0x0000000000000000000000000000000000b447b980aea5accb5fd68789f6b099
     * amount: BigInt | number
     * returns:
     * 64 byte buffer
     */

  }, {
    key: "sigHashBuff",
    value: function sigHashBuff(utxoId, amount) {
      var valueBuf = _ethereumjsUtil["default"].setLengthLeft(Buffer.from(_ethereumjsUtil["default"].padToEven(amount.toString(16)), 'hex'), 32);

      var utxoIdBuf = Buffer.from(utxoId.replace('0x', ''), 'hex');
      return Buffer.concat([utxoIdBuf, valueBuf]);
    }
  }, {
    key: "bufferToBytes32Array",
    value: function bufferToBytes32Array(buffer) {
      var output = [];
      var offset = 0;

      for (var i = 0; i < buffer.length / 32; i++) {
        var bytes32 = "0x".concat(buffer.slice(offset, offset + 32).toString('hex'));
        offset += 32;
        output.push(bytes32);
      }

      return output;
    }
  }, {
    key: "txFromProof",
    value: function txFromProof(proof) {
      var tx = _transaction["default"].fromRaw(this.parseTxDataFromProof(proof));

      return tx;
    }
  }, {
    key: "parseTxDataFromProof",
    value: function parseTxDataFromProof(proof) {
      var wordLength = 32;
      var buf = Buffer.alloc(proof.length * wordLength);

      for (var i = 0; i < proof.length; i++) {
        Buffer.from(proof[i].replace('0x', ''), 'hex').copy(buf, i * wordLength, 0, wordLength);
      }

      var offset = buf.readUInt8(32);
      var length = buf.readUInt16BE(34);
      var txData = Buffer.alloc(length);
      buf.copy(txData, 0, offset, offset + length);
      return txData;
    }
  }, {
    key: "fastSellUTXO",
    value: function fastSellUTXO(utxo, plasmaChain, rootChain, marketMakerUrl, signer) {
      var promiEvent = (0, _web3CorePromievent["default"])();
      plasmaChain.getConfig().then(function (nodeConfig) {
        var inputs = [new _input["default"](utxo.outpoint)];
        var outputs = [new _output["default"](utxo.output.value, nodeConfig.exitHandlerAddr, utxo.output.color)];

        var exitingUtxoTransfer = _transaction["default"].transfer(inputs, outputs);

        return Exit.fastSell(exitingUtxoTransfer, plasmaChain, rootChain, marketMakerUrl, promiEvent, signer);
      });
      return promiEvent.eventEmitter;
    }
  }, {
    key: "fastSellAmount",
    value: function fastSellAmount(account, amount, color, plasmaChain, rootChain, marketMakerUrl, signer) {
      var promiEvent = (0, _web3CorePromievent["default"])();
      Promise.all([plasmaChain.getUnspent(account, color), plasmaChain.getConfig()]).then(function (_ref) {
        var _ref2 = _slicedToArray(_ref, 2),
            unspent = _ref2[0],
            nodeConfig = _ref2[1];

        var exitingUtxoTransfer = _transaction["default"].transferFromUtxos(unspent, account, nodeConfig.exitHandlerAddr, amount, color);

        return Exit.fastSell(exitingUtxoTransfer, plasmaChain, rootChain, marketMakerUrl, promiEvent, signer);
      });
      return promiEvent.eventEmitter;
    }
  }, {
    key: "fastSell",
    value: function fastSell(spendTx, plasmaChain, rootChain, marketMakerUrl, _promiEvent, signer) {
      var amount = (0, _jsbiUtils.bi)(spendTx.outputs[0].value);
      var fastSellRequest = {};

      var promiEvent = _promiEvent || (0, _web3CorePromievent["default"])();

      (signer ? signer.signTx(spendTx) : spendTx.signWeb3(rootChain)).then(function (signedTx) {
        return (0, _helpers.sendSignedTransaction)(plasmaChain, signedTx.hex());
      }).then(function (transferTx) {
        var tx = _transaction["default"].fromRaw(transferTx.raw);

        var utxoId = new _outpoint["default"](tx.hash(), 0).getUtxoId();
        var sigHashBuff = Exit.sigHashBuff(utxoId, amount);
        fastSellRequest.tx = transferTx;
        fastSellRequest.sigHashBuff = sigHashBuff;

        var _Period$periodBlockRa = _period["default"].periodBlockRange(transferTx.blockNumber);

        var _Period$periodBlockRa2 = _slicedToArray(_Period$periodBlockRa, 2);

        fastSellRequest.effectiveBlock = _Period$periodBlockRa2[1];
        return plasmaChain.eth.getTransaction(_ethereumjsUtil["default"].bufferToHex(tx.inputs[0].prevout.hash));
      }).then(function (inputTx) {
        fastSellRequest.inputTx = inputTx;
        promiEvent.eventEmitter.emit('transfer', fastSellRequest);
        return Exit.signAndSendFastSellRequest(fastSellRequest, rootChain, marketMakerUrl, signer);
      }).then(promiEvent.resolve)["catch"](promiEvent.reject);
      return promiEvent.eventEmitter;
    }
  }, {
    key: "signAndSendFastSellRequest",
    value: function signAndSendFastSellRequest(fastSellRequest, rootChain, marketMakerUrl, signer) {
      var sigHash = "0x".concat(fastSellRequest.sigHashBuff.toString('hex'));
      return (signer ? signer.signMessage(sigHash) : _transaction["default"].signMessageWithWeb3(rootChain, sigHash)).then(function (sig) {
        var vBuff = Buffer.alloc(32);
        vBuff.writeInt8(sig.v, 31);
        var signedData = Exit.bufferToBytes32Array(Buffer.concat([fastSellRequest.sigHashBuff, Buffer.from(sig.r), Buffer.from(sig.s), vBuff]));
        fastSellRequest.signedData = signedData;
        return (0, _nodeFetch["default"])(marketMakerUrl, {
          method: "POST",
          mode: "cors",
          cache: "no-cache",
          headers: {
            "Content-Type": "application/json"
          },
          body: JSON.stringify(fastSellRequest)
        }).then(function (response) {
          return response.json();
        });
      });
    }
  }]);

  return Exit;
}();

exports["default"] = Exit;
module.exports = exports.default;