"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.extendWeb3 = extendWeb3;
exports.calcInputs = calcInputs;
exports.calcOutputs = calcOutputs;
exports.periodBlockRange = periodBlockRange;
exports.getTxWithYoungestBlock = getTxWithYoungestBlock;
exports.getYoungestInputTx = getYoungestInputTx;
exports.getProof = getProof;
exports.sendSignedTransaction = sendSignedTransaction;
exports.consolidateUTXOs = consolidateUTXOs;
exports.simulateSpendCond = simulateSpendCond;

var _ethereumjsUtil = require("ethereumjs-util");

var _output = _interopRequireDefault(require("./output"));

var _outpoint = _interopRequireDefault(require("./outpoint"));

var _period = _interopRequireDefault(require("./period"));

var _transaction = _interopRequireDefault(require("./transaction"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance"); }

function _iterableToArrayLimit(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var fromRaw = function fromRaw(u) {
  return {
    output: u.output,
    outpoint: _outpoint["default"].fromRaw(u.outpoint)
  };
};

function extendWeb3(web3Instance) {
  // `_extend` for web3 0.2x.x, `extend` for 1.x
  var extend = web3Instance._extend || web3Instance.extend; // eslint-disable-line no-underscore-dangle, max-len

  var getUnspent = {
    name: 'getUnspent',
    call: 'plasma_unspent',
    params: 2,
    inputFormatters: [extend.formatters.inputAddressFormatter],
    outputFormatter: function outputFormatter(unspent) {
      if (Array.isArray(unspent)) {
        // web3 0.2x.x passes in an array
        return unspent.map(fromRaw);
      }

      return fromRaw(unspent);
    }
  };
  extend({
    methods: [new extend.Method(_objectSpread({}, getUnspent, {
      name: '_getUnspent'
    })), new extend.Method(_objectSpread({}, getUnspent, {
      params: 1,
      name: 'getUnspentByAddress'
    })), new extend.Method(_objectSpread({}, getUnspent, {
      params: 0,
      name: 'getUnspentAll'
    })), new extend.Method({
      name: 'getColor',
      call: 'plasma_getColor',
      params: 1,
      inputFormatters: [extend.formatters.inputAddressFormatter],
      outputFormatter: Number
    }), new extend.Method({
      name: '_getColors',
      call: 'plasma_getColors',
      params: 2,
      inputFormatters: [function (a) {
        return a;
      }, function (a) {
        return a;
      }],
      outputFormatter: String
    }), new extend.Method({
      name: 'status',
      call: 'plasma_status',
      params: 0,
      inputFormatters: [],
      outputFormatter: String
    }), new extend.Method({
      name: 'getConfig',
      call: 'plasma_getConfig',
      params: 0,
      inputFormatters: [],
      outputFormatter: function outputFormatter(a) {
        return a;
      }
    }), new extend.Method({
      name: 'getValidatorInfo',
      call: 'validator_getAddress',
      params: 0,
      inputFormatters: [],
      outputFormatter: function outputFormatter(a) {
        return a;
      }
    }), new extend.Method({
      name: 'checkSpendingCondition',
      call: 'checkSpendingCondition',
      params: 1,
      inputFormatters: [function (tx) {
        return tx.hex();
      }],
      outputFormatter: function outputFormatter(_ref) {
        var error = _ref.error,
            outputs = _ref.outputs;
        return {
          error: error,
          outputs: outputs.map(_output["default"].fromJSON)
        };
      }
    }), new extend.Method({
      name: 'getPeriodByBlockHeight',
      call: 'plasma_getPeriodByBlockHeight',
      params: 1,
      inputFormatters: [function (a) {
        return a;
      }],
      outputFormatter: function outputFormatter(a) {
        return a;
      }
    })]
  });

  web3Instance.getUnspent = function () {
    var _ref2;

    var hasCb = typeof (_ref2 = arguments.length - 1, _ref2 < 0 || arguments.length <= _ref2 ? undefined : arguments[_ref2]) === 'function';
    var paramCount = hasCb ? arguments.length - 1 : arguments.length;

    if (paramCount === 0) {
      return web3Instance.getUnspentAll();
    }

    if (paramCount === 1) {
      return web3Instance.getUnspentByAddress.apply(web3Instance, arguments);
    }

    return web3Instance._getUnspent.apply(web3Instance, arguments); // eslint-disable-line no-underscore-dangle
  };
  /* eslint-disable no-underscore-dangle */


  web3Instance.getColors = function (type) {
    var typeStr = (type || '').toLowerCase();

    if (typeStr === 'erc721' || typeStr === 'nft') {
      return web3Instance._getColors(true, null);
    }

    if (typeStr === 'erc1948' || type === 'nst') {
      return web3Instance._getColors(false, true);
    }

    if (typeStr === 'erc20') {
      return web3Instance._getColors(false, false);
    }

    return Promise.all([web3Instance._getColors(false, false), web3Instance._getColors(true, false), web3Instance._getColors(false, true)]).then(function (_ref3) {
      var _ref4 = _slicedToArray(_ref3, 3),
          erc20 = _ref4[0],
          erc721 = _ref4[1],
          erc1948 = _ref4[2];

      return {
        erc20: erc20,
        erc721: erc721,
        erc1948: erc1948
      };
    });
  };
  /* eslint-enable no-underscore-dangle */


  return web3Instance;
}
/**
 * DEPRECATED: Use Transaction.calcInputs instead
 */


function calcInputs(unspent, from, amount, color) {
  console.warn('DEPRECATED: use Tx.calcInputs instead'); // eslint-disable-line no-console

  return _transaction["default"].calcInputs(unspent, from, amount, color);
}
/**
 * DEPRECATED: Use Transaction.calcOutputs instead
 */


function calcOutputs(unspent, inputs, from, to, amount, color) {
  console.warn('DEPRECATED: use Tx.calcOutputs instead'); // eslint-disable-line no-console

  return _transaction["default"].calcOutputs(unspent, inputs, from, to, amount, color);
}
/**
 * DEPRECATED: Use Period.periodBlockRange instead
 */


function periodBlockRange(blockNumber) {
  console.warn('DEPRECATED: use Period.periodBlockRange instead'); // eslint-disable-line no-console

  return _period["default"].periodBlockRange(blockNumber);
}
/**
 * Finds the youngest tx in a given array.
 *
 * Youngest tx is the one with biggest block number.
 * @param {LeapTransaction[]} txs
 * @returns {InputTx} youngest tx and its index
 */


function getTxWithYoungestBlock(txs) {
  return txs.reduce(function (res, tx, i) {
    if (tx.blockNumber > res.tx.blockNumber) {
      res.index = i;
      res.tx = tx;
    }

    return res;
  }, {
    index: 0,
    tx: txs[0]
  });
}
/**
 * Returns the youngest input for a given tx.
 *
 * Youngest input is the one which references tx with biggest block number.
 * @param {ExtendedWeb3|LeapProvider} plasma instance of Leap Web3
 * @param {Tx} tx
 * @returns {Promise<InputTx>} promise that resolves to youngest input tx and its index
 */


function getYoungestInputTx(plasma, tx) {
  return Promise.all(tx.inputs.map(function (i) {
    return (plasma.eth || plasma).getTransaction((0, _ethereumjsUtil.bufferToHex)(i.prevout.hash));
  })).then(getTxWithYoungestBlock);
}
/**
 * Creates proof of period inclusion for a given tx
 *
 * @param {ExtendedWeb3|LeapProvider} plasma instance of Leap Web3
 * @param {LeapTransaction} tx
 * @param {PeriodOpts} periodOpts — options for cosntructed Period as defined in Period constructor
 * @returns {Promise<Proof>} promise that resolves to period inclusion proof
 */


function getProof(plasma, tx) {
  var periodOpts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  return plasma.getPeriodByBlockHeight(tx.blockNumber).then(function (periodData) {
    if (periodData) {
      Object.assign(periodOpts, {
        validatorData: periodData
      });
    } else {
      var msg = "No period data for the given tx. Height: ".concat(tx.blockNumber);

      if (!periodOpts.validatorData) {
        throw new Error(msg);
      } else {
        console.warn(msg, 'Using fallback values'); // eslint-disable-line no-console
      }
    }

    return _period["default"].periodForTx(plasma, tx, periodOpts);
  }).then(function (period) {
    return period.proof(_transaction["default"].fromRaw(tx.raw));
  });
}
/**
 * Polls the node for transaction receipt 50 times with 100 ms interval (5 seconds in total).
 * Resolves to tx receipt if it is available, otherwise resolves to nothing.
 */


var awaitForReceipt = function awaitForReceipt(txHash, promise, plasma, round) {
  if (round >= 50) return Promise.reject(new Error('Transaction not included in block after 5 secs.'));
  return promise.then(function (receipt) {
    if (receipt && receipt.blockHash) {
      return Promise.resolve(receipt);
    }

    var next = new Promise(function (resolve) {
      return setTimeout(function () {
        return (plasma.eth || plasma).getTransaction(txHash).then(resolve);
      }, 100);
    });
    return awaitForReceipt(txHash, next, plasma, round + 1);
  });
};

var send = function send(plasma, txHex) {
  if (plasma.currentProvider) {
    // Web3
    return new Promise(function (resolve, reject) {
      plasma.currentProvider.send({
        jsonrpc: '2.0',
        id: 42,
        method: 'eth_sendRawTransaction',
        params: [txHex]
      }, function (err, res) {
        if (err) {
          return reject(err);
        }

        return resolve(res.result);
      });
    });
  }

  return plasma.send('eth_sendRawTransaction', [txHex]); // Ethers.js
};
/**
 * Sends a signed transaction to the node
 *
 * @param {ExtendedWeb3} plasma instance of Leap Web3 or JSONRPCProvider of Ethers.js
 * @param {string} txHex - transaction to send
 * @returns {Promise<Receipt>} promise that resolves to a transaction receipt
 */


function sendSignedTransaction(plasma, txHex) {
  // eslint-disable-next-line no-console
  console.warn('DEPRECATED: use Web3#sendSignedTransaction or LeapProvider#sendTransaction instead');
  return send(plasma, txHex).then(function (resp) {
    return awaitForReceipt(resp, Promise.resolve(), plasma, 0);
  });
}
/**
 * DEPRECATED: use Tx.consolidateUTXOs instead
 */


function consolidateUTXOs(utxos) {
  console.warn('DEPRECATED: use Tx.consolidateUTXOs instead'); // eslint-disable-line no-console

  return _transaction["default"].consolidateUTXOs(utxos);
}
/**
 * Returns outputs for spending condition tx
 *
 * @param {ExtendedWeb3 | LeapProvider} plasma
 * @param {Tx} tx
 *
 * @returns {Promise<Array<Output>>}
 */


function simulateSpendCond(plasma, tx) {
  return plasma.checkSpendingCondition(tx);
}