"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = exports.MAX_COLOR = exports.NST_OUT_LENGTH = exports.OUT_LENGTH = void 0;

var _assert = _interopRequireDefault(require("assert"));

var _jsbiUtils = require("jsbi-utils");

var _util = _interopRequireDefault(require("./util"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

// 32 bytes - value, 20 bytes - address, 2 bytes color
var OUT_LENGTH = 54; // 32 - data, + normal output like above

exports.OUT_LENGTH = OUT_LENGTH;
var NST_OUT_LENGTH = OUT_LENGTH + 32;
exports.NST_OUT_LENGTH = NST_OUT_LENGTH;
var MAX_COLOR = Math.pow(2, 16);
exports.MAX_COLOR = MAX_COLOR;

var valueHex = function valueHex(value) {
  return (0, _jsbiUtils.BigInt)(value).toString(16);
};

var Output =
/*#__PURE__*/
function () {
  function Output(valueOrObject, address, color, data) {
    _classCallCheck(this, Output);

    if (_typeof(valueOrObject) === 'object' && !(0, _jsbiUtils.isBigInt)(valueOrObject)) {
      _util["default"].ensureSafeValue(valueOrObject.value);

      this.value = (0, _jsbiUtils.BigInt)(valueOrObject.value);
      this.color = valueOrObject.color || 0;
      this.address = valueOrObject.address;

      if (this.isNST()) {
        this.data = valueOrObject.data;
      }
    } else {
      _util["default"].ensureSafeValue(valueOrObject);

      this.value = (0, _jsbiUtils.BigInt)(valueOrObject);
      this.address = address;
      this.color = color || 0;

      if (this.isNST()) {
        this.data = data;
      }
    }

    if (!this.isNFT() && !this.isNST() && (0, _jsbiUtils.lessThan)(this.value, (0, _jsbiUtils.BigInt)(1))) {
      throw new Error('Output value is < 1');
    }

    if (this.color < 0 || this.color > MAX_COLOR) {
      throw new Error("Color out of range (0\u2013".concat(MAX_COLOR, ")"));
    }

    if (this.isNST() && !_util["default"].isBytes32(this.data)) {
      throw new Error('data is not a 32 bytes hex string');
    }
  }

  _createClass(Output, [{
    key: "isNFT",
    value: function isNFT() {
      return _util["default"].isNFT(this.color);
    }
  }, {
    key: "isNST",
    value: function isNST() {
      return _util["default"].isNST(this.color);
    }
    /* eslint-disable class-methods-use-this */

  }, {
    key: "getSize",
    value: function getSize() {
      // transfer output
      return this.isNST() ? NST_OUT_LENGTH : OUT_LENGTH;
    }
    /* eslint-enable class-methods-use-this */

  }, {
    key: "toJSON",
    value: function toJSON() {
      var rsp = {
        address: this.address,
        value: this.value.toString(10),
        color: this.color
      };

      if (this.isNST()) {
        rsp.data = this.data;
      }

      return rsp;
    }
    /**
     * Instantiate output from json object.
     * @param {Object} json
     * @returns {Outpoint}
     */

  }, {
    key: "toRaw",
    value: function toRaw() {
      var dataBuf = Buffer.alloc(this.getSize());
      dataBuf.write(valueHex(this.value).padStart(64, '0'), 0, 32, 'hex');
      dataBuf.writeUInt16BE(this.color, 32);
      dataBuf.write(this.address.replace('0x', ''), 34, 'hex');

      if (this.isNST()) {
        dataBuf.write(this.data.replace('0x', ''), 54, 'hex');
      }

      return dataBuf;
    }
  }], [{
    key: "fromJSON",
    value: function fromJSON(json) {
      (0, _assert["default"])(json, 'Output data is required.');
      return new Output(json);
    }
    /**
     * Instantiate output from serialized data.
     * @param {Buffer} data
     * @param {offset} offset to read from in buffer
     * @returns {Output}
     */

  }, {
    key: "fromRaw",
    value: function fromRaw(buf) {
      var offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
      var color = buf.readUInt16BE(offset + 32);

      var valueString = _util["default"].toHexString(buf.slice(offset, offset + 32));

      var value = (0, _jsbiUtils.BigInt)(valueString);

      var address = _util["default"].toHexString(buf.slice(offset + 34, offset + 54));

      var data;

      if (_util["default"].isNST(color)) {
        data = _util["default"].toHexString(buf.slice(offset + 54, offset + 54 + 32));
      } // transfer output


      return new Output(value, address, color, data);
    }
  }]);

  return Output;
}();

exports["default"] = Output;