"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _ethereumjsUtil = _interopRequireDefault(require("ethereumjs-util"));

var _jsbiUtils = require("jsbi-utils");

var _input = _interopRequireWildcard(require("./input"));

var _output = _interopRequireWildcard(require("./output"));

var _outpoint = _interopRequireDefault(require("./outpoint"));

var _util = _interopRequireDefault(require("./util"));

var _type = _interopRequireDefault(require("./type"));

var _constants = require("./constants");

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = Object.defineProperty && Object.getOwnPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : {}; if (desc.get || desc.set) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } } newObj["default"] = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance"); }

function _iterableToArrayLimit(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var EMPTY_BUF = Buffer.alloc(32, 0);

var Transaction =
/*#__PURE__*/
function () {
  function Transaction(type) {
    var inputs = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var outputs = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];
    var options = arguments.length > 3 ? arguments[3] : undefined;

    _classCallCheck(this, Transaction);

    this.type = type;
    this.inputs = inputs;
    this.outputs = outputs;
    this.options = options; // recover signer if we have signatures without signer

    if (this.inputs.length && this.inputs[0].v && !this.inputs[0].signer) {
      this.recoverTxSigner();
    }
  }

  _createClass(Transaction, [{
    key: "getSize",

    /*
    * Returns raw transaction size.
    * See `toRaw` for details.
    */
    value: function getSize() {
      if (this.type === _type["default"].DEPOSIT) {
        var output = this.outputs[0];
        return output.getSize() + 6;
      }

      if (this.type === _type["default"].TRANSFER) {
        return 2 + this.inputs.reduce(function (s, i) {
          return s + i.getSize();
        }, 0) + this.outputs.reduce(function (s, o) {
          return s + o.getSize();
        }, 0);
      }

      return this.toRaw().length;
    } // Recovers signer address for each of the inputs
    // Requires inputs to be signed.

  }, {
    key: "recoverTxSigner",
    value: function recoverTxSigner() {
      var _this = this;

      this.inputs.map(function (i) {
        return i.recoverSigner(_this.sigDataBuf());
      });
    } // Returns sigData as Buffer.
    // Calculated as follows:
    // 1. serialize to bytes
    // 2. strip out input signatures

  }, {
    key: "sigDataBuf",
    value: function sigDataBuf() {
      return Transaction.sigDataBufStatic(this.type, this.toRaw(), this.inputs.length);
    } // Returns sigData as hex string

  }, {
    key: "sigData",
    value: function sigData() {
      return _ethereumjsUtil["default"].bufferToHex(this.sigDataBuf());
    } // Signs each input with provided private keys
    // @param {Array} privKeys - array of private keys strings.

  }, {
    key: "sign",
    value: function sign(privKeys) {
      if (this.type === _type["default"].PERIOD_VOTE) {
        var sig = _ethereumjsUtil["default"].ecsign(this.inputs[0].prevout.hash, // Buffer.from(this.options.merkleRoot.replace('0x', ''), 'hex'),
        Buffer.from(privKeys[0].replace('0x', ''), 'hex'));

        this.inputs[0].setSig(sig.r, sig.s, sig.v, // sig
        _ethereumjsUtil["default"].bufferToHex(_ethereumjsUtil["default"].privateToAddress(privKeys[0])) // signer
        );
        return this;
      }

      if (privKeys.length !== this.inputs.length) {
        throw Error('amount of private keys doesn\'t match amount of inputs');
      }

      var startIdx = this.type === _type["default"].TRANSFER ? 0 : 1;

      for (var i = startIdx; i < privKeys.length; i++) {
        if (privKeys[i]) {
          var sigHashBuf = _ethereumjsUtil["default"].hashPersonalMessage(this.sigDataBuf());

          var _sig = _ethereumjsUtil["default"].ecsign(sigHashBuf, Buffer.from(privKeys[i].replace('0x', ''), 'hex'));

          this.inputs[i].setSig(_sig.r, _sig.s, _sig.v, // sig
          _ethereumjsUtil["default"].bufferToHex(_ethereumjsUtil["default"].privateToAddress(privKeys[i])) // signer
          );
        }
      }

      return this;
    }
  }, {
    key: "signAll",
    value: function signAll(privKey) {
      return this.sign(this.inputs.map(function () {
        return privKey;
      }));
    }
  }, {
    key: "getConditionSig",
    value: function getConditionSig(privKey) {
      if (this.type !== _type["default"].SPEND_COND) {
        throw Error("invalid Tx type: ".concat(this.type));
      }

      var sigHash = Buffer.alloc(32, 0);
      this.sigHashBuf().copy(sigHash, 12, 0, 20);
      return _ethereumjsUtil["default"].ecsign(sigHash, Buffer.from(privKey.replace('0x', ''), 'hex'));
    }
  }, {
    key: "sigHash",
    value: function sigHash() {
      return _ethereumjsUtil["default"].bufferToHex(this.sigHashBuf());
    }
  }, {
    key: "sigHashBuf",
    value: function sigHashBuf() {
      var sigDataBuf = this.sigDataBuf();

      if (this.type === _type["default"].SPEND_COND) {
        return _ethereumjsUtil["default"].ripemd160(_ethereumjsUtil["default"].keccak256(sigDataBuf));
      }

      return _ethereumjsUtil["default"].keccak256(sigDataBuf);
    }
  }, {
    key: "signWeb3",
    value: function signWeb3(web3) {
      var _this2 = this;

      var accountIndex = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

      if (this.type === _type["default"].PERIOD_VOTE) {
        return Transaction.signMessageWithWeb3(web3, this.inputs[0].prevout.hash, accountIndex).then(function (_ref) {
          var r = _ref.r,
              s = _ref.s,
              v = _ref.v,
              signer = _ref.signer;

          _this2.inputs[0].setSig(r, s, v, signer);

          return _this2;
        });
      }

      return Transaction.signMessageWithWeb3(web3, this.sigData(), accountIndex).then(function (_ref2) {
        var r = _ref2.r,
            s = _ref2.s,
            v = _ref2.v,
            signer = _ref2.signer;
        var inputs = _this2.type === _type["default"].TRANSFER ? _this2.inputs : _this2.inputs.slice(1);
        inputs.forEach(function (input) {
          input.setSig(r, s, v, signer);
        });
        return _this2;
      });
    } // Returns tx hash as Buffer

  }, {
    key: "hashBuf",
    value: function hashBuf() {
      var raw = this.toRaw();

      if (!this.isSigned()) {
        throw Error('not signed yet');
      }

      return _ethereumjsUtil["default"].keccak256(raw);
    }
  }, {
    key: "isSigned",
    value: function isSigned() {
      return this.type !== _type["default"].TRANSFER || !this.toRaw().slice(35, 67).equals(EMPTY_BUF);
    } // Returns tx hash as hex string

  }, {
    key: "hash",
    value: function hash() {
      return _ethereumjsUtil["default"].bufferToHex(this.hashBuf());
    } // Returns serialized tx bytes as hex string

  }, {
    key: "hex",
    value: function hex() {
      return _ethereumjsUtil["default"].bufferToHex(this.toRaw());
    } // Checks if this tx is equal to `another`

  }, {
    key: "equals",
    value: function equals(another) {
      return this.toRaw().equals(another.toRaw());
    }
    /**
     * Returns perceived "value" of the transaction and it's color.
     *
     * If tx has outputs, the first outputs is considered as a value bearing.
     * For Exit transactions, returns the value of the connected output of
     * the previous tx (if provided).
     *
     * @param {Transaction} prevTx (optional) previous tx connected to the first input's outpoint
     * @returns {object} { value, color }
     */

  }, {
    key: "value",
    value: function value() {
      var prevTx = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

      // assuming first output is transfer, second one is change
      if (this.outputs && this.outputs.length > 0) {
        var _this$outputs$ = this.outputs[0],
            value = _this$outputs$.value,
            color = _this$outputs$.color;
        return {
          value: value,
          color: color
        };
      }

      if (this.type === _type["default"].EXIT && prevTx) {
        var _prevTx$outputs$this$ = prevTx.outputs[this.inputs[0].prevout.index],
            _value = _prevTx$outputs$this$.value,
            _color = _prevTx$outputs$this$.color;
        return {
          value: _value,
          color: _color
        };
      }

      return {
        value: (0, _jsbiUtils.BigInt)(0),
        color: 0
      };
    }
    /**
     * Returns perceived "from" address.
     *
     * If tx has inputs, the first input is considered as a source.
     * Address is either:
     * 1. address of the connected output of the previous tx (if provided)
     * 2. signer of the first input (if tx is signed)
     *
     * Returns null address, if neither prevTx provided nor tx is signed.
     *
     * @param {Transaction} prevTx (optional) previous tx connected to the first input's outpoint
     * @returns {string} address
     */

  }, {
    key: "from",
    value: function from() {
      var prevTx = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

      if (!this.inputs || !this.inputs.length) {
        return _constants.EMPTY_ADDRESS;
      }

      if (prevTx && prevTx.hashBuf().equals(this.inputs[0].prevout.hash)) {
        return prevTx.outputs[this.inputs[0].prevout.index].address;
      }

      return this.inputs[0].signer || _constants.EMPTY_ADDRESS;
    }
    /**
       * Returns perceived "to" address.
       *
       * If tx has outputs, the first outputs is considered as a destination.   *
       * Returns null address, if tx has no outputs
       *
       * @returns {string} address
       */

  }, {
    key: "to",
    value: function to() {
      if (!this.outputs || !this.outputs.length) {
        return _constants.EMPTY_ADDRESS;
      }

      return this.outputs[0].address;
    }
    /**
    *
    * Serialized this tx in bytes as follows:
    *
    *
    * Deposit (60 bytes):
    *
    * 0 -   1 byte type
    * 1 -   4 bits number of inputs (always 1)
    * 1 -   4 bits number of outputs (always 1)
    * 2 -   4 bytes depositId
    * 6 -  32 bytes value
    * 38 -  2 bytes color
    * 40 - 20 bytes address
    * 60
    *
    * Transfer (314 bytes):
    *
    * 0 -   1 byte type
    * 1 -   4 bits number of inputs
    * 1 -   4 bits number of outputs
    * 2 -   32 bytes prev tx          $
    * 34 -  1 byte output pos         $      first input
    * 35 -  32 bytes r-signature        #
    * 67 -  32 bytes s-signature        #    first signature
    * 99 -  1 byte v-signature          #
    * 108 - 32 bytes prev tx          $
    * 140 - 1 byte output pos         $      second input
    * 141 - 32 bytes r-signature        #
    * 173 - 32 bytes s-signature        #    second signature
    * 205 - 1 byte v-signature          #
    * 206 - 32 bytes value            $
    * 238 - 2 bytes color             $
    * 240 - 20 bytes address          $      first output
    * 260 - 32 bytes value              #
    * 292 - 2 bytes color               #
    * 294 - 20 bytes address            #    second output
    * 314
    *
    * */

    /* eslint-disable prefer-destructuring */

  }, {
    key: "toRaw",
    value: function toRaw() {
      var payload;
      var inputs = [];

      var writeLengths = function writeLengths(inputsLength, outputsLength) {
        // write ins and outs length as nibbles
        if (inputsLength > 15) {
          throw new Error('Too many inputs (>15)');
        }

        if (outputsLength > 15) {
          throw new Error('Too many outputs (>15)');
        }

        payload.writeUInt8(16 * inputsLength + outputsLength, 1);
      };

      if (this.type === _type["default"].DEPOSIT) {
        payload = Buffer.alloc(6, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(1, 1);
        payload.writeUInt32BE(this.options.depositId, 2);
      } else if (this.type === _type["default"].EPOCH_LENGTH_V1) {
        payload = Buffer.alloc(6, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(0, 0);
        payload.writeUInt32BE(this.options.epochLength, 2);
      } else if (this.type === _type["default"].EPOCH_LENGTH_V2) {
        payload = Buffer.alloc(10, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(0, 0);
        payload.writeUInt32BE(this.options.epochLength, 2);
        payload.writeUInt32BE(this.options.blockHeight, 6);
      } else if (this.type === _type["default"].MIN_GAS_PRICE) {
        payload = Buffer.alloc(10, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(0, 0);
        var valueHex = (0, _jsbiUtils.BigInt)(this.options.minGasPrice).toString(16);
        payload.write(valueHex.padStart(16, '0'), 2, 8, 'hex');
      } else if (this.type === _type["default"].VALIDATOR_JOIN) {
        payload = Buffer.alloc(60, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(0, 0);
        payload.writeUInt16BE(this.options.slotId, 2);
        payload.write(this.options.tenderKey.replace('0x', ''), 4, 'hex');
        payload.writeUInt32BE(this.options.eventsCount, 36);
        payload.write(this.options.signerAddr.replace('0x', ''), 40, 'hex');
      } else if (this.type === _type["default"].VALIDATOR_LOGOUT) {
        payload = Buffer.alloc(64, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(0, 0);
        payload.writeUInt16BE(this.options.slotId, 2);
        payload.write(this.options.tenderKey.replace('0x', ''), 4, 'hex');
        payload.writeUInt32BE(this.options.eventsCount, 36);
        payload.writeUInt32BE(this.options.activationEpoch, 40);
        payload.write(this.options.newSigner.replace('0x', ''), 44, 'hex');
      } else if (this.type === _type["default"].PERIOD_VOTE) {
        payload = Buffer.alloc(3, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(1, 0);
        payload.writeUInt8(this.options.slotId, 2);
        inputs = this.inputs;
      } else if (this.type === _type["default"].EXIT) {
        payload = Buffer.alloc(34, 0);
        payload.writeUInt8(this.type, 0);

        _util["default"].arrayToRaw(this.inputs).copy(payload, 1, 0, 33);
      } else if (this.type === _type["default"].TRANSFER) {
        payload = Buffer.alloc(2, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(this.inputs.length, this.outputs.length);
        inputs = this.inputs;
      } else if (this.type === _type["default"].SPEND_COND) {
        payload = Buffer.alloc(2, 0);
        payload.writeUInt8(this.type, 0);
        writeLengths(this.inputs.length, this.outputs.length);
        inputs = this.inputs;
      } else {
        payload = Buffer.alloc(1, 0);
        payload.writeUInt8(this.type, 0);
        inputs = this.inputs;
      }

      return Buffer.concat([payload, _util["default"].arrayToRaw(inputs), _util["default"].arrayToRaw(this.outputs)]);
    }
    /* eslint-enable prefer-destructuring */

  }, {
    key: "toJSON",
    value: function toJSON() {
      var prevTx = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;
      var json = {
        type: this.type,
        inputs: this.inputs.map(function (inp) {
          return inp.toJSON();
        }),
        outputs: this.outputs.map(function (out) {
          return out.toJSON();
        })
      };

      if (this.type === _type["default"].TRANSFER) {
        json.to = this.to();
        json.from = this.from(prevTx);

        var _this$value = this.value(prevTx),
            value = _this$value.value,
            color = _this$value.color;

        json.value = value.toString(10);
        json.color = color;
      }

      if (this.isSigned()) {
        json.hash = this.hash();
      }

      if (this.options) {
        json.options = this.options;
      }

      return json;
    }
  }], [{
    key: "epochLengthV1",
    value: function epochLengthV1(epochLength) {
      // eslint-disable-next-line no-console
      console.warn('DEPRECATED: use epochLength(epochLength, blockHeight) instead');
      return new Transaction(_type["default"].EPOCH_LENGTH_V1, [], [], {
        epochLength: epochLength
      });
    }
  }, {
    key: "epochLength",
    value: function epochLength(_epochLength, blockHeight) {
      if (!(_epochLength > 0)) {
        throw new Error('Invalid epoch length (> 0)');
      }

      if (!(blockHeight > 0)) {
        throw new Error('Invalid block height (> 0)');
      }

      return new Transaction(_type["default"].EPOCH_LENGTH_V2, [], [], {
        epochLength: _epochLength,
        blockHeight: blockHeight
      });
    }
  }, {
    key: "minGasPrice",
    value: function minGasPrice(_minGasPrice) {
      _util["default"].ensureSafeValue(_minGasPrice);

      return new Transaction(_type["default"].MIN_GAS_PRICE, [], [], {
        minGasPrice: (0, _jsbiUtils.BigInt)(_minGasPrice)
      });
    }
  }, {
    key: "validatorJoin",
    value: function validatorJoin(slotId, tenderKey, eventsCount, signerAddr) {
      return new Transaction(_type["default"].VALIDATOR_JOIN, [], [], {
        slotId: slotId,
        eventsCount: eventsCount,
        signerAddr: signerAddr.toLowerCase(),
        tenderKey: tenderKey.toLowerCase()
      });
    }
  }, {
    key: "validatorLogout",
    value: function validatorLogout(slotId, tenderKey, eventsCount, activationEpoch, newSigner) {
      return new Transaction(_type["default"].VALIDATOR_LOGOUT, [], [], {
        slotId: slotId,
        eventsCount: eventsCount,
        tenderKey: tenderKey.toLowerCase(),
        activationEpoch: activationEpoch,
        newSigner: newSigner.toLowerCase()
      });
    }
  }, {
    key: "periodVote",
    value: function periodVote(slotId, input) {
      return new Transaction(_type["default"].PERIOD_VOTE, [input], [], {
        slotId: slotId
      });
    }
  }, {
    key: "deposit",
    value: function deposit(depositId, value, address, color, data) {
      return new Transaction(_type["default"].DEPOSIT, [], [new _output["default"](value, address, color, data)], {
        depositId: depositId
      });
    }
  }, {
    key: "exit",
    value: function exit(input) {
      return new Transaction(_type["default"].EXIT, [input]);
    }
  }, {
    key: "transfer",
    value: function transfer(inputs, outputs) {
      return new Transaction(_type["default"].TRANSFER, inputs, outputs);
    }
  }, {
    key: "transferFromUtxos",
    value: function transferFromUtxos(utxos, from, to, value, color) {
      var inputs = Transaction.calcInputs(utxos, from, value, color);
      var outputs = Transaction.calcOutputs(utxos, inputs, from, to, value, color);
      return this.transfer(inputs, outputs);
    }
  }, {
    key: "spendCond",
    value: function spendCond(inputs, outputs) {
      inputs[0].type = _type["default"].SPEND_COND; // eslint-disable-line no-param-reassign

      return new Transaction(_type["default"].SPEND_COND, inputs, outputs);
    }
  }, {
    key: "sigDataBufStatic",
    value: function sigDataBufStatic(type, raw, inputsLength) {
      var noSigs = Buffer.alloc(raw.length, 0);

      if (type === _type["default"].TRANSFER || type === _type["default"].PERIOD_VOTE) {
        var offset = 2; // copy type and lengths

        raw.copy(noSigs, 0, 0, offset);

        for (var i = 0; i < inputsLength; i += 1) {
          raw.copy(noSigs, offset, offset, offset + 33);
          offset += type !== _type["default"].TRANSFER && i === 0 ? 33 : 98;
        }

        raw.copy(noSigs, offset, offset, raw.length);
      } else if (type === _type["default"].SPEND_COND) {
        // spend_cond sighash is calculated with 00 for msgData
        // as message data will contain the signature for spending the tx
        var bytesStripped = raw.readUInt16BE(35); // offset of msgData length in 1st input

        noSigs = Buffer.alloc(raw.length - bytesStripped, 0);
        var readOffset = 0;
        var writeOffset = 0; // copy type and lengths

        raw.copy(noSigs, writeOffset, readOffset, 2);
        readOffset += 2;
        writeOffset += 2;

        for (var _i = 0; _i < inputsLength; _i += 1) {
          raw.copy(noSigs, writeOffset, readOffset, readOffset + 33);

          if (_i === 0) {
            var msgLength = raw.readUInt16BE(readOffset + 33);
            readOffset += 33 + 2 + msgLength;
            writeOffset += 33 + 2;
            var scriptLength = raw.readUInt16BE(readOffset);
            raw.copy(noSigs, writeOffset, readOffset, readOffset + 2 + scriptLength);
            readOffset += 2 + scriptLength;
            writeOffset += 2 + scriptLength;
          } else {
            // all but first inputs have 65 bytes for sig
            readOffset += 33 + 65;
            writeOffset += 33 + 65;
          }
        }

        raw.copy(noSigs, writeOffset, readOffset, raw.length);
      }

      return noSigs;
    }
  }, {
    key: "signMessageWithWeb3",
    value: function signMessageWithWeb3(web3, message) {
      var accountIndex = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
      var version = typeof web3.version === 'string' ? web3.version : web3.version.api;
      return new Promise(function (resolve, reject) {
        web3.eth.getAccounts(function (err, accounts) {
          if (err) {
            return reject(new Error("getAccounts err: ".concat(err)));
          }

          if (!accounts[accountIndex]) {
            return reject(new Error("No account at index ".concat(accountIndex)));
          }

          var args = version.startsWith('0') ? [accounts[accountIndex], message] : [message, accounts[accountIndex]];
          args.push(function (sigErr, sig) {
            if (sigErr) {
              return reject(sigErr);
            }

            var _ethUtil$fromRpcSig = _ethereumjsUtil["default"].fromRpcSig(sig),
                r = _ethUtil$fromRpcSig.r,
                s = _ethUtil$fromRpcSig.s,
                v = _ethUtil$fromRpcSig.v;

            return resolve({
              r: r,
              s: s,
              v: v,
              signer: accounts[accountIndex]
            });
          }); // const { sign } = web3.currentProvider.isMetaMask ? web3.eth.personal : web3.eth;

          var sign = web3.eth.personal.sign;
          return sign.apply(void 0, args);
        });
      });
    }
  }, {
    key: "fromJSON",
    value: function fromJSON(_ref3) {
      var type = _ref3.type,
          inputs = _ref3.inputs,
          outputs = _ref3.outputs,
          options = _ref3.options;
      return new Transaction(type, inputs.map(_input["default"].fromJSON), outputs.map(_output["default"].fromJSON), options);
    }
  }, {
    key: "parseToParams",
    value: function parseToParams(transaction) {
      var bufs = this.parseToBuf(transaction);
      return bufs.parts.map(function (buf) {
        return "0x".concat(buf.toString('hex'));
      });
    } // Constructs Transaction from given raw bytes
    // @returns {Transaction}

  }, {
    key: "fromRaw",
    value: function fromRaw(transaction) {
      var dataBuf = transaction;

      if (!Buffer.isBuffer(transaction)) {
        var dataHex = transaction.replace('0x', '');
        dataBuf = Buffer.alloc(dataHex.length / 2);
        dataBuf.write(dataHex, 'hex');
      }

      var type = dataBuf.readUInt8(0);

      switch (type) {
        case _type["default"].DEPOSIT:
          {
            if (dataBuf.length !== _output.OUT_LENGTH + 6 && dataBuf.length !== _output.NST_OUT_LENGTH + 6) {
              throw new Error('malformed deposit tx.');
            }

            var depositId = dataBuf.readUInt32BE(2);

            var output = _output["default"].fromRaw(dataBuf.slice(6));

            return new Transaction(_type["default"].DEPOSIT, [], [output], {
              depositId: depositId
            });
          }

        case _type["default"].EPOCH_LENGTH_V1:
          {
            if (dataBuf.length !== 6) {
              throw new Error('malformed epoch tx.');
            }

            var epochLength = dataBuf.readUInt32BE(2);
            return Transaction.epochLengthV1(epochLength);
          }

        case _type["default"].EPOCH_LENGTH_V2:
          {
            if (dataBuf.length !== 10) {
              throw new Error('malformed epoch tx.');
            }

            var _epochLength2 = dataBuf.readUInt32BE(2);

            var blockHeight = dataBuf.readUInt32BE(6);
            return Transaction.epochLength(_epochLength2, blockHeight);
          }

        case _type["default"].MIN_GAS_PRICE:
          {
            if (dataBuf.length !== 10) {
              throw new Error('malformed gas price tx.');
            }

            var gasString = _util["default"].toHexString(dataBuf.slice(2, 10));

            var minGasPrice = (0, _jsbiUtils.BigInt)(gasString);
            return Transaction.minGasPrice(minGasPrice);
          }

        case _type["default"].VALIDATOR_JOIN:
          {
            if (dataBuf.length !== 60) {
              throw new Error('malformed validatorJoin tx.');
            }

            var slotId = dataBuf.readUInt16BE(2);
            var tenderKey = "0x".concat(dataBuf.slice(4, 36).toString('hex'));
            var eventsCount = dataBuf.readUInt32BE(36);
            var signerAddr = "0x".concat(dataBuf.slice(40, 60).toString('hex'));
            return Transaction.validatorJoin(slotId, tenderKey, eventsCount, signerAddr);
          }

        case _type["default"].VALIDATOR_LOGOUT:
          {
            if (dataBuf.length !== 64) {
              throw new Error('malformed validatorLogout tx.');
            }

            var _slotId = dataBuf.readUInt16BE(2);

            var _tenderKey = "0x".concat(dataBuf.slice(4, 36).toString('hex'));

            var _eventsCount = dataBuf.readUInt32BE(36);

            var activationEpoch = dataBuf.readUInt32BE(40);
            var newSigner = "0x".concat(dataBuf.slice(44, 64).toString('hex'));
            return Transaction.validatorLogout(_slotId, _tenderKey, _eventsCount, activationEpoch, newSigner);
          }

        case _type["default"].PERIOD_VOTE:
          {
            if (dataBuf.length !== 101) {
              throw new Error('malformed periodVote tx.');
            }

            var _slotId2 = dataBuf.readUInt8(2);

            var ins = [];
            var sigHashBuf = dataBuf.slice(3, 35);
            ins.push(_input["default"].fromRaw(dataBuf, 3, sigHashBuf));
            return new Transaction(_type["default"].PERIOD_VOTE, ins, [], {
              slotId: _slotId2
            });
          }

        case _type["default"].EXIT:
          {
            if (dataBuf.length !== 34) {
              throw new Error('malformed exit tx.');
            }

            var outpoint = _outpoint["default"].fromRaw(dataBuf, 1);

            return new Transaction(_type["default"].EXIT, [new _input["default"](outpoint)], []);
          }

        case _type["default"].TRANSFER:
          {
            var insOuts = dataBuf.readUInt8(1);
            var insLength = insOuts >> 4; // eslint-disable-line no-bitwise

            var outsLength = insOuts & 0xF; // eslint-disable-line no-bitwise

            var _ins = [];

            var _sigHashBuf = _ethereumjsUtil["default"].hashPersonalMessage(Transaction.sigDataBufStatic(type, dataBuf, insLength));

            for (var i = 0; i < insLength; i += 1) {
              _ins.push(_input["default"].fromRaw(dataBuf, 2 + i * _input.SPEND_INPUT_LENGTH, _sigHashBuf));
            }

            var outs = [];
            var offset = 2 + insLength * _input.SPEND_INPUT_LENGTH;

            for (var _i2 = 0; _i2 < outsLength; _i2 += 1) {
              var out = _output["default"].fromRaw(dataBuf, offset);

              offset += out.getSize();
              outs.push(out);
            }

            return new Transaction(type, _ins, outs);
          }

        case _type["default"].SPEND_COND:
          {
            var _insOuts = dataBuf.readUInt8(1);

            var _insLength = _insOuts >> 4; // eslint-disable-line no-bitwise


            var _outsLength = _insOuts & 0xF; // eslint-disable-line no-bitwise


            var _ins2 = [];
            var _offset = 2;

            var _sigHashBuf2 = _ethereumjsUtil["default"].hashPersonalMessage(Transaction.sigDataBufStatic(type, dataBuf, _insLength));

            for (var _i3 = 0; _i3 < _insLength; _i3 += 1) {
              var input = void 0;

              if (_i3 === 0) {
                input = _input["default"].fromRaw(dataBuf, _offset, _i3 === 0 ? _type["default"].SPEND_COND : 0);
              } else {
                input = _input["default"].fromRaw(dataBuf, _offset, _sigHashBuf2);
              }

              _ins2.push(input);

              _offset += input.getSize();
            }

            var _outs = [];

            for (var _i4 = 0; _i4 < _outsLength; _i4 += 1) {
              var _out = _output["default"].fromRaw(dataBuf, _offset);

              _offset += _out.getSize();

              _outs.push(_out);
            }

            return new Transaction(_type["default"].SPEND_COND, _ins2, _outs);
          }

        default:
          {
            throw new Error("unknown transaction type: ".concat(type, "."));
          }
      }
    }
  }, {
    key: "calcInputs",
    value: function calcInputs(unspent, from, amount) {
      var color = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 0;
      var limit = arguments.length > 4 ? arguments[4] : undefined;
      var myUnspent = unspent.filter(function (_ref4) {
        var output = _ref4.output;
        return output.color === color && output.address.toLowerCase() === from.toLowerCase();
      });
      var exact = myUnspent.find(function (utxo) {
        return (0, _jsbiUtils.equal)((0, _jsbiUtils.BigInt)(utxo.output.value), (0, _jsbiUtils.BigInt)(amount));
      });
      if (exact) return [new _input["default"](exact.outpoint)];
      var inputs = [];
      var sum = (0, _jsbiUtils.BigInt)(0);

      for (var i = 0; i < myUnspent.length; i += 1) {
        inputs.push(new _input["default"](myUnspent[i].outpoint));
        sum = (0, _jsbiUtils.add)(sum, (0, _jsbiUtils.BigInt)(myUnspent[i].output.value));

        if ((0, _jsbiUtils.greaterThanOrEqual)(sum, (0, _jsbiUtils.BigInt)(amount))) {
          break;
        }

        if (limit && inputs.length + 1 === limit) {
          var _ret = function () {
            var diff = (0, _jsbiUtils.subtract)((0, _jsbiUtils.BigInt)(amount), sum);
            var oneMore = myUnspent.slice(i + 1).find(function (u) {
              return (0, _jsbiUtils.greaterThanOrEqual)((0, _jsbiUtils.BigInt)(u.output.value), diff);
            });

            if (oneMore) {
              inputs.push(new _input["default"](oneMore.outpoint));
              sum = (0, _jsbiUtils.add)(sum, (0, _jsbiUtils.BigInt)(oneMore.output.value));
            }

            return "break";
          }();

          if (_ret === "break") break;
        }
      }

      if ((0, _jsbiUtils.lessThan)(sum, (0, _jsbiUtils.BigInt)(amount))) {
        throw new Error('Not enough inputs');
      }

      return inputs;
    } // ToDo: handle inputs from different accounts
    // ToDo: handle different input colors

  }, {
    key: "calcOutputs",
    value: function calcOutputs(unspent, inputs, from, to, amount, color) {
      if (unspent.length === 0) {
        throw new Error('Unspent is empty');
      }

      var inInputs = function inInputs(u) {
        return inputs.findIndex(function (input) {
          return u.outpoint.equals(input.prevout);
        }) > -1;
      };

      var inputUtxos = unspent.filter(inInputs);

      if (_util["default"].isNFT(color) || _util["default"].isNST(color)) {
        return inputUtxos.map(function (i) {
          return new _output["default"](i.output.value, to, color, i.output.data);
        });
      }

      var sum = inputUtxos.reduce(function (a, u) {
        return (0, _jsbiUtils.add)(a, (0, _jsbiUtils.BigInt)(u.output.value));
      }, (0, _jsbiUtils.BigInt)(0));

      if ((0, _jsbiUtils.lessThan)(sum, (0, _jsbiUtils.BigInt)(amount))) {
        throw new Error('Not enough inputs');
      }

      var outputs = [new _output["default"](amount, to.toLowerCase(), color)];

      if ((0, _jsbiUtils.greaterThan)(sum, (0, _jsbiUtils.BigInt)(amount))) {
        outputs.push(new _output["default"]((0, _jsbiUtils.subtract)(sum, (0, _jsbiUtils.BigInt)(amount)), from.toLowerCase(), color));
      }

      return outputs;
    }
    /**
     * Creates an array of consolidation transactions for a given UTXO list. Each tx with up to 15 inputs.
     *
     * @param {Unspent[]} utxos - UTXOs to consolidate
     * @returns {LeapTransaction[]} array of consolidate transactions
     */

  }, {
    key: "consolidateUTXOs",
    value: function consolidateUTXOs(utxos) {
      if (utxos.length === 0) {
        return [];
      }

      var colors = Array.from(new Set(utxos.map(function (u) {
        return u.output.color;
      })));
      var addrs = Array.from(new Set(utxos.map(function (u) {
        return u.output.address;
      })));

      if (colors.length > 1) {
        throw new Error("Expected UTXOs only for one color, got ".concat(colors.length));
      }

      if (addrs.length > 1) {
        throw new Error("Expected UTXOs only for one address, got ".concat(addrs.length));
      }

      var _colors = _slicedToArray(colors, 1),
          color = _colors[0];

      var _addrs = _slicedToArray(addrs, 1),
          address = _addrs[0];

      var chunks = [[]];
      utxos.forEach(function (utxo, i) {
        var currentChunk = chunks[chunks.length - 1];
        currentChunk.push(utxo);

        if (currentChunk.length === 15 && i !== utxos.length - 1) {
          chunks.push([]);
        }
      });
      return chunks.map(function (chunk) {
        var inputs = chunk.map(function (u) {
          return new _input["default"](u.outpoint);
        });
        var value = chunk.reduce(function (v, u) {
          return (0, _jsbiUtils.add)(v, (0, _jsbiUtils.bi)(u.output.value));
        }, _jsbiUtils.ZERO);
        return Transaction.transfer(inputs, [new _output["default"](value, address, Number(color))]);
      });
    }
  }]);

  return Transaction;
}();

exports["default"] = Transaction;
module.exports = exports.default;