"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _assert = _interopRequireDefault(require("assert"));

var _jsbiUtils = require("jsbi-utils");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var Util =
/*#__PURE__*/
function () {
  function Util() {
    _classCallCheck(this, Util);
  }

  _createClass(Util, null, [{
    key: "isUint",

    /**
     * Test whether an object is a uint.
     * @param {Number?} value
     * @returns {Boolean}
     */
    value: function isUint(value) {
      return Util.isInt(value) && value >= 0;
    }
    /**
     * Test whether an object is a uint8.
     * @param {Number?} value
     * @returns {Boolean}
     */

  }, {
    key: "isU8",
    value: function isU8(value) {
      return (value & 0xff) === value; // eslint-disable-line no-bitwise
    }
    /**
     * Test whether an object is a uint16.
     * @param {Number?} value
     * @returns {Boolean}
     */

  }, {
    key: "isU16",
    value: function isU16(value) {
      return (value & 0xffff) === value; // eslint-disable-line no-bitwise
    }
    /**
     * Test whether an object is a uint32.
     * @param {Number?} value
     * @returns {Boolean}
     */

  }, {
    key: "isU32",
    value: function isU32(value) {
      return value >>> 0 === value; // eslint-disable-line no-bitwise
    }
    /**
     * Test whether an object is a uint53.
     * @param {Number?} value
     * @returns {Boolean}
     */

  }, {
    key: "isU64",
    value: function isU64(value) {
      return Util.isUint(value);
    }
    /**
     * Test whether a string is a plain
     * ascii string (no control characters).
     * @param {String} str
     * @returns {Boolean}
     */

    /**
     * Find index of a buffer in an array of buffers.
     * @param {Buffer[]} items
     * @param {Buffer} data - Target buffer to find.
     * @returns {Number} Index (-1 if not found).
     */

  }, {
    key: "indexOf",
    value: function indexOf(items, data) {
      (0, _assert["default"])(Array.isArray(items));
      (0, _assert["default"])(Buffer.isBuffer(data));

      for (var i = 0; i < items.length; i++) {
        var item = items[i];
        (0, _assert["default"])(Buffer.isBuffer(item));

        if (item.equals(data)) {
          return i;
        }
      }

      return -1;
    }
    /**
     * Test to see if a string starts with a prefix.
     * @param {String} str
     * @param {String} prefix
     * @returns {Boolean}
     */

  }, {
    key: "startsWith",
    value: function startsWith(str, prefix) {
      (0, _assert["default"])(typeof str === 'string');

      if (!str.startsWith) {
        return str.indexOf(prefix) === 0;
      }

      return str.startsWith(prefix);
    }
    /**
     * Test whether a string is hex (length must be even).
     * Note that this _could_ await a false positive on
     * base58 strings.
     * @param {String?} str
     * @returns {Boolean}
     */

  }, {
    key: "isBytes",
    value: function isBytes(str) {
      if (typeof str !== 'string') {
        return false;
      }

      var trunk = str.indexOf('0x') === 0 ? str.substring(2, str.length) : str;
      return str.length % 2 === 0 && /^[0-9A-Fa-f]+$/.test(trunk);
    }
    /**
     * Test whether an object is a 256 bit hash (hex string).
     * @param {String?} hash
     * @returns {Boolean}
     */

  }, {
    key: "isBytes32",
    value: function isBytes32(hash) {
      if (typeof hash !== 'string') {
        return false;
      }

      var trunk = hash.indexOf('0x') === 0 ? hash.substring(2, hash.length) : hash;
      return trunk.length === 64 && Util.isBytes(trunk);
    }
  }, {
    key: "writeUint64",
    value: function writeUint64(buff, value) {
      var offset = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
      var big = ~~(value / 0xFFFFFFFF); // eslint-disable-line no-bitwise

      buff.writeUInt32BE(big, offset);
      var low = value % 0xFFFFFFFF - big;
      buff.writeUInt32BE(low, offset + 4);
    }
  }, {
    key: "readUint64",
    value: function readUint64(buff) {
      var offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
      return (0, _jsbiUtils.BigInt)(Util.toHexString(buff.slice(offset, offset + 8)), 16);
    }
  }, {
    key: "arrayToRaw",
    value: function arrayToRaw(arr) {
      // todo: possible performance improvement if totalLength supplied to Buffer.concat
      return Buffer.concat(arr.map(function (v) {
        return v.toRaw();
      }));
    }
  }, {
    key: "toHexString",
    value: function toHexString(buffer) {
      return "0x".concat(buffer.toString('hex'));
    }
    /**
     * Returns an array of numbers in the given interval.
     *
     * @param {Number} from left bound of the interval
     * @param {Number} to right bound of the interval
     * @returns {Array}
     */

  }, {
    key: "range",
    value: function range(from, to) {
      return Array.from(new Array(to - from + 1), function (_, i) {
        return i + from;
      });
    }
  }, {
    key: "isNFT",
    value: function isNFT(color) {
      return color > Math.pow(2, 15) && color <= Math.pow(2, 15) + Math.pow(2, 14);
    }
  }, {
    key: "isNST",
    value: function isNST(color) {
      return color > Math.pow(2, 15) + Math.pow(2, 14);
    }
  }, {
    key: "ensureSafeValue",
    value: function ensureSafeValue(value) {
      if (typeof value === 'number' && value > Number.MAX_SAFE_INTEGER) {
        throw new Error("Use JSBI.BigInt or string for large values. Using Number could lead to precision loss");
      }
    }
  }]);

  return Util;
}();

exports["default"] = Util;
module.exports = exports.default;